//============================================================================//
// Copyright 2014 Alachisoft.  All rights reserved.                    //
//============================================================================//

using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using DBSync.DataSource;
using DBSync.DataObject;
using Alachisoft.NCache.Web.Caching;
using Alachisoft.NCache.Runtime.Dependencies;

namespace DBSync.UI
{
    public partial class AppForm1 : Form
    {
        private string                      connString;
        private SqlDatasource               sqlDataSource;
        private OtherDatasource             oleDbDataSource;

        #region Delegates
        public delegate void UpdateLvwDelegate(Product p);
        public delegate void LoadCacheItemDelegate(Product p);
        #endregion

        // holds the products data
        private Product[]                   products;

        // number of records to fetch from database
        private int                         numOfRows;

        /// <summary>         /// Cache Instance ///         /// </summary>
        public static Cache _cache;

        private string                      dbms;

        public AppForm1()
        {
            InitializeCache();
            InitializeComponent();
            
            dbms = System.Configuration.ConfigurationManager.AppSettings["dbms"];

            InitializeDB();

            numOfRows = 15;
            _cache.ItemRemoved += new CacheItemRemovedCallback(Cache_ItemRemoved);
            _cache.ItemUpdated += new CacheItemUpdatedCallback(Cache_ItemUpdated);

            // load data at startup
            btnLoad_Click(this, EventArgs.Empty);
            this.Visible = true;
        }

        #region >>      === UI Event Handler(s) ===        <<

        private void btnLoad_Click(object sender, EventArgs e)
        {
            LoadAllFromSource();
            UpdateListView();
            AddAllInCache();
        }

        private void btnDone_Click(object sender, EventArgs e)
        {
            this.Dispose(true);
            this.Close();
        }

        #endregion

        #region >>      === ListView Operation(s) ===       <<

        /// <summary>
        /// Update the unit price and the color of the expired item.
        /// </summary>
        /// <param name="product"></param>
        private void UpdateListViewItem(Product product)
        {
            ListViewItem lvItem = this.lvwSyncItems.Items[product.ProductID - 1];//.Find(product.ProductID.ToString(), false);

            if (lvItem != null)
            {
                int oldPrice = Convert.ToInt32(lvItem.SubItems[2].Text);
                lvItem.SubItems[2].Text = product.UnitPrice.ToString();

                if (oldPrice < product.UnitPrice)
                {
                    lvItem.BackColor = Color.Red;
                }
                else if (oldPrice > product.UnitPrice)
                {
                    lvItem.BackColor = Color.Green;
                }
                else
                {
                    lvItem.BackColor = Color.Aqua;
                }
            }
            else
                MessageBox.Show("Item not found in the cache");
        }


        private void UpdateListView()
        {
            lvwSyncItems.Items.Clear();
            for (int i = 0; i < products.Length; i++)
            {
                string[] str = { products[i].ProductID.ToString(), products[i].ProductName, products[i].UnitPrice.ToString() };
                ListViewItem lvItem = new ListViewItem(str, -1);
                lvwSyncItems.Items.Add(lvItem);
            }
        }

        #endregion

        #region >>      === DataBase Operation(s) ===       <<

        /// <summary>Establishes the DB Connection
        /// </summary>
        /// <returns>'true' if connection is successfull and 'false' if connection fails</returns>
        private bool InitializeDB()
        {
            if (!dbms.Equals("other", StringComparison.OrdinalIgnoreCase))
            {
                sqlDataSource = new SqlDatasource();
                connString = System.Configuration.ConfigurationManager.AppSettings["conn-string"];

                sqlDataSource.Connect(connString);
            }
            else
            {
                oleDbDataSource = new OtherDatasource();
                connString = System.Configuration.ConfigurationManager.AppSettings["conn-string"];
                oleDbDataSource.Connect(connString);
            }
            return true;
        }

        /// <summary>
        /// Releases the connection.
        /// </summary>
        public void DisconnectDB()
        {
            if (sqlDataSource != null)
            {
                sqlDataSource.DisConnect();
                sqlDataSource = null;
            }

            if (oleDbDataSource != null)
            {
                oleDbDataSource.DisConnect();
                oleDbDataSource = null;
            }
        }

        /// <summary>
        /// Loads all the products from the database.
        /// </summary>
        private void LoadAllFromSource()
        {
            if (!dbms.Equals("other", StringComparison.OrdinalIgnoreCase))
                products = sqlDataSource.LoadProducts(numOfRows);
            else
                products = oleDbDataSource.LoadProducts(numOfRows);
        }

        /// <summary>
        /// Get the product values from the database.
        /// </summary>
        /// <param name="key">ProductID</param>
        /// <returns></returns>
        private Product LoadItemFromSource(string key)
        {
            Product product = null;
            if (!dbms.Equals("other", StringComparison.OrdinalIgnoreCase))
            {
                product = (Product)sqlDataSource.LoadProduct(key);
            }
            else
            {
                product = (Product)oleDbDataSource.LoadProduct(key);
            }
            return product;
        }

        #endregion 

        #region >>      === Cache Operation(s) ===      <<

        /// <summary>
        /// Add the products in the cache.
        /// </summary>
        public void AddAllInCache()
        {
            try
            {
                for (int i = 0; i < numOfRows; i++)
                {
                    AddItemInCache(products[i]);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }


        private void AddItemInCache(Product prod)
        {
            try
            {
                CacheItem item = new CacheItem(prod);
                if (dbms.Equals("sql2005", StringComparison.OrdinalIgnoreCase))
                {
                    // for sql server 2005
                    item.Dependency = new SqlCacheDependency(connString, "select ProductID from dbo.products where ProductID = " + prod.ProductID.ToString());
                    _cache.Insert(prod.ProductID.ToString(), item,DSWriteOption.None,null);
                }
                else if (dbms.Equals("sql2000", StringComparison.OrdinalIgnoreCase))
                {
                    // for sql server 2000
                    item.Dependency = DBDependencyFactory.CreateSqlCacheDependency(connString, prod.ProductID.ToString());

                    // Note: To change this cache Key format you'll also have to change the trigger appropriatly in the database.
                    _cache.Insert(prod.ProductID.ToString() + ":dbo.Products", item, DSWriteOption.None, null);
                }
                else if (dbms.Equals("other", StringComparison.OrdinalIgnoreCase))
                {
                    // for any other database (e.g oracle) 
                    item.Dependency = DBDependencyFactory.CreateOleDbCacheDependency(connString, prod.ProductID.ToString());
                    _cache.Insert(prod.ProductID.ToString() + ":Products", item, DSWriteOption.None, null);
                }
            }
            catch (Exception ex)
            {
            }
        }

        /// <summary>
        /// Initializes the cache.
        /// </summary>
        private void InitializeCache()
        {
            try
            {
               _cache = NCache.InitializeCache(System.Configuration.ConfigurationManager.AppSettings["cache-id"].ToString());
               _cache.Clear();
            }
            catch (Exception e)
            {
                throw e;
            }
        }


        /// <summary>
        /// Dispose Cache.
        /// </summary>
        private void DisposeCache()
        {
           _cache.Dispose();
        }

        #endregion

        #region >>      === Cache Notification Handler(s) ===     <<
        /// <summary>
        /// Notification handler. invoked when an item is removed from cache.
        /// </summary>
        /// <param name="key"></param>
        /// <param name="value"></param>
        /// <param name="reason"></param>
        void Cache_ItemRemoved(object key, object value, CacheItemRemovedReason reason)
        {
            if (key != null)
            {
                Product prod = (Product)LoadItemFromSource(((Product)value).ProductID.ToString());
                if (prod != null)
                {
                    this.Invoke(new UpdateLvwDelegate(UpdateListViewItem), new object[] { prod });
                    this.Invoke(new LoadCacheItemDelegate(AddItemInCache), new object[] { prod });
                }
            }
        }

        /// <summary>
        /// Called when item is updated in the cache.
        /// </summary>
        /// <param name="key"></param>
        void Cache_ItemUpdated(object key)
        {
            if (key != null)
            {
                Product prod = (Product)LoadItemFromSource(key.ToString());

                if (prod != null)
                {
                    this.Invoke(new UpdateLvwDelegate(UpdateListViewItem), new object[] { prod });
                    this.Invoke(new LoadCacheItemDelegate(AddItemInCache), new object[] { prod });
                }
            }
        }

        #endregion 

    }
}